@@ -1,8 +1,10 @@
-__version__ = '1.0.7'
+__version__ = '1.0.8'
 
 from . import msda
 from . import utils
+from . import anamoly
 
 # To get sub-modules
 from .msda import *
 from .utils import *
+from .anamoly import *
